"""Training/evaluation utilities module.

This module contains utilities and tools used to instantiate training sessions. It also contains
the prediction consumer interface used by metrics and loggers to receive iteration data during
training. See :mod:`thelper.optim.metrics` for more information on metrics.
"""

import json
import logging
import os
from typing import Any, AnyStr, Dict, List, Optional, Union  # noqa: F401

import cv2 as cv
import numpy as np
import sklearn.metrics
import torch

import thelper.concepts
import thelper.ifaces
import thelper.typedefs  # noqa: F401
import thelper.utils
from thelper.ifaces import ClassNamesHandler, FormatHandler, PredictionConsumer
from thelper.optim.eval import compute_iou
from thelper.tasks.detect import BoundingBox

logger = logging.getLogger(__name__)


class PredictionCallback(PredictionConsumer):
    """Callback function wrapper compatible with the consumer interface.

    This interface is used to hide user-defined callbacks into the list of prediction consumers given
    to trainer implementations. The callbacks must always be compatible with the list of arguments
    defined by ``thelper.typedefs.IterCallbackParams``, but may also receive extra arguments defined in
    advance and passed to the constructor of this class.

    Attributes:
        callback_func: user-defined function to call on every update from the trainer.
        callback_kwargs: user-defined extra arguments to provide to the callback function.
    """

    def __init__(self, callback_func, callback_kwargs=None):
        # type: (thelper.typedefs.IterCallbackType, thelper.typedefs.IterCallbackParams) -> None
        assert callback_func is not None and \
            (isinstance(callback_func, str) or callable(callback_func)), \
            "invalid callback function, must be importable string or callable object"
        if isinstance(callback_func, str):
            callback_func = thelper.utils.import_function(callback_func)
        thelper.utils.check_func_signature(callback_func, thelper.typedefs.IterCallbackParams)
        assert callback_kwargs is None or \
            (isinstance(callback_kwargs, dict) and
             not any([p in callback_kwargs for p in thelper.typedefs.IterCallbackParams])), \
            "invalid callback kwargs (must be dict, and not contain overlap with default args)"
        self.callback_func = callback_func
        self.callback_kwargs = callback_kwargs

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this consumer."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(callback_func={repr(self.callback_func)}, callback_kwargs={repr(self.callback_kwargs)})"

    def update(self, *args, **kwargs):
        """Forwards the latest prediction data from the training session to the user callback."""
        return self.callback_func(*args, **kwargs, **self.callback_kwargs)


@thelper.concepts.classification
class ClassifLogger(PredictionConsumer, ClassNamesHandler, FormatHandler):
    """Classification output logger.

    This class provides a simple logging interface for accumulating and saving the predictions of a classifier.
    By default, all predictions will be logged. However, a confidence threshold can be set to focus on "hard"
    samples if necessary. It also optionally offers tensorboardX-compatible output images that can be saved
    locally or posted to tensorboard for browser-based visualization.

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example consumer; it is used for lookup/printing only
            "logger": {
                # this type is used to instantiate the confusion matrix report object
                "type": "thelper.train.utils.ClassifLogger",
                "params": {
                    # log the three 'best' predictions for each sample
                    "top_k": 3,
                    # keep updating a set of 10 samples for visualization via tensorboardX
                    "viz_count": 10
                }
            },
            # ...
        }
        # ...

    Attributes:
        top_k: number of 'best' predictions to keep for each sample (along with the gt label).
        conf_threshold: threshold used to eliminate uncertain predictions.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        target_name: name of the targeted label (may be 'None' if all classes are used).
        target_idx: index of the targeted label (may be 'None' if all classes are used).
        viz_count: number of tensorboardX images to generate and update at each epoch.
        report_count: number of samples to print in reports (use 'None' if all samples must be printed).
        log_keys: list of metadata field keys to copy from samples into the log for each prediction.
        force_softmax: specifies whether a softmax operation should be applied to the prediction scores
            obtained from the trainer.
        score: array used to store prediction scores for logging.
        true: array used to store groundtruth labels for logging.
        meta: array used to store metadata pulled from samples for logging.
        format: output format of the produced log (supports: text, CSV)
    """

    def __init__(self,
                 top_k=1,               # type: int
                 conf_threshold=None,   # type: Optional[thelper.typedefs.Number]
                 class_names=None,      # type: Optional[List[AnyStr]]
                 target_name=None,      # type: Optional[AnyStr]
                 viz_count=0,           # type: int
                 report_count=None,     # type: Optional[int]
                 log_keys=None,         # type: Optional[List[AnyStr]]
                 force_softmax=True,    # type: bool
                 format=None,           # type: Optional[AnyStr]
                 ):                     # type: (...) -> None
        """Receives the logging parameters & the optional class label names used to decorate the log."""
        assert isinstance(top_k, int) and top_k > 0, "invalid top-k value"
        assert conf_threshold is None or (isinstance(conf_threshold, (float, int)) and 0 < conf_threshold <= 1), \
            "classification confidence threshold should be 'None' or float in ]0, 1]"
        assert isinstance(viz_count, int) and viz_count >= 0, "invalid image count to visualize"
        assert report_count is None or (isinstance(report_count, int) and report_count >= 0), \
            "invalid report sample count"
        assert log_keys is None or isinstance(log_keys, list), "invalid list of sample keys to log"
        self.top_k = top_k
        self.target_name = target_name
        self.target_idx = None
        self.conf_threshold = conf_threshold
        self.viz_count = viz_count
        self.report_count = report_count
        self.log_keys = log_keys if log_keys is not None else []
        self.force_softmax = force_softmax
        self.score = None
        self.true = None
        self.meta = None
        ClassNamesHandler.__init__(self, class_names)
        FormatHandler.__init__(self, format)

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this consumer."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(top_k={repr(self.top_k)}, conf_threshold={repr(self.conf_threshold)}, " + \
            f"class_names={repr(self.class_names)}, target_name={repr(self.target_name)}, " + \
            f"viz_count={repr(self.viz_count)}, report_count={repr(self.report_count)}, " + \
            f"log_keys={repr(self.log_keys)}, force_softmax={repr(self.force_softmax)})"

    @ClassNamesHandler.class_names.setter
    def class_names(self, class_names):
        """Sets the class label names that must be predicted by the model."""
        ClassNamesHandler.class_names.fset(self, class_names)
        if self.target_name is not None and self.class_names is not None:
            assert self.target_name in self.class_indices, \
                f"could not find target name {repr(self.target_name)} in class names list"
            self.target_idx = self.class_indices[self.target_name]
        else:
            self.target_idx = None

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert isinstance(task, thelper.tasks.Classification), "classif report only impl for classif tasks"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to update function"
        if self.score is None or self.score.size != max_iters:
            self.score = np.asarray([None] * max_iters)
            self.true = np.asarray([None] * max_iters)
            self.meta = {key: np.asarray([None] * max_iters) for key in self.log_keys}
        if task.class_names != self.class_names:
            self.class_names = task.class_names
        assert pred.dim() == 2, "current classif logger impl only supports 2D outputs (BxC)"
        assert pred.shape[1] == len(self.class_names), "unexpected prediction class dimension size"
        if target is None or target.numel() == 0:
            self.true[iter_idx] = None
        else:
            assert target.dim() == 1, "gt should be batched (1D) tensor"
            assert pred.shape[0] == target.shape[0], "prediction/gt tensors batch size mismatch"
            self.true[iter_idx] = target.numpy()
        if self.force_softmax:
            with torch.no_grad():
                pred = torch.nn.functional.softmax(pred, dim=1)
        self.score[iter_idx] = pred.numpy()
        for meta_key in self.log_keys:
            assert meta_key in sample, f"could not extract sample field with key {repr(meta_key)}"
            val = sample[meta_key]
            assert isinstance(val, (list, np.ndarray, torch.Tensor)), f"field {repr(meta_key)} should be batched"
            self.meta[meta_key][iter_idx] = val if isinstance(val, list) else val.tolist()

    def render(self):
        """Returns an image of predicted outputs as a numpy-compatible RGBA image drawn by pyplot."""
        if self.viz_count == 0:
            return None
        if self.score is None or self.true is None:
            return None
        raise NotImplementedError  # TODO

    def report_text(self):
        # type: () -> Optional[AnyStr]
        return self.report_csv()

    def report_csv(self):
        # type: () -> Optional[AnyStr]
        """Returns the logged metadata of predicted samples.

        The returned object is a print-friendly CSV string that can be consumed directly by tensorboardX. Note
        that this string might be very long if the dataset is large (i.e. it will contain one line per sample).
        """
        if isinstance(self.report_count, int) and self.report_count <= 0:
            return None
        if self.score is None or self.true is None:
            return None
        pack = list(zip(*[(*pack, )
                          for packs in zip(self.score, self.true, *self.meta.values())
                          for pack in zip(*packs)]))
        logdata = {key: np.stack(val, axis=0) for key, val in zip(["pred", "target", *self.meta.keys()], pack)}
        assert all([len(val) == len(logdata["target"]) for val in logdata.values()]), "messed up unpacking"
        header = "target_name,target_score"
        for k in range(self.top_k):
            header += f",pred_{k + 1}_name,pred_{k + 1}_score"
        for meta_key in self.log_keys:
            header += f",{str(meta_key)}"
        lines = []
        for sample_idx in range(len(logdata["target"])):
            gt_label_idx = int(logdata["target"][sample_idx])
            pred_scores = logdata["pred"][sample_idx]
            sorted_score_idxs = np.argsort(pred_scores)[::-1]
            sorted_scores = pred_scores[sorted_score_idxs]
            if self.conf_threshold is None or gt_label_idx is None or \
                    pred_scores[gt_label_idx] >= self.conf_threshold:
                if gt_label_idx is not None:
                    entry = f"{self.class_names[gt_label_idx]},{pred_scores[gt_label_idx]:2.4f}"
                else:
                    entry = f"<unknown>,{0.0:2.4f}"
                for k in range(self.top_k):
                    entry += f",{self.class_names[sorted_score_idxs[k]]},{sorted_scores[k]:2.4f}"
                for meta_key in self.log_keys:
                    entry += f",{str(logdata[meta_key][sample_idx])}"
                lines.append(entry)
                if isinstance(self.report_count, int) and len(lines) >= self.report_count:
                    break
        return "\n".join([header, *lines])

    def reset(self):
        """Toggles a reset of the internal state, emptying storage arrays."""
        self.score = None
        self.true = None
        self.meta = None


@thelper.concepts.classification
class ClassifReport(PredictionConsumer, ClassNamesHandler, FormatHandler):
    """Classification report interface.

    This class provides a simple interface to ``sklearn.metrics.classification_report`` so that all
    count-based metrics can be reported at once under a string-based representation.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example consumer; it is used for lookup/printing only
            "report": {
                # this type is used to instantiate the classification report object
                "type": "thelper.train.utils.ClassifReport",
                # we do not need to provide any parameters to the constructor, defaults are fine
                "params": {
                    # optional parameter that will indicate output as JSON is desired, plain 'text' otherwise
                    "format": "json"
                }
            },
            # ...
        }
        # ...

    Attributes:
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        pred: queue used to store the top-1 (best) predicted class indices at each iteration.
        format: output format of the produced log (supports: text, JSON)
    """

    def __init__(self, class_names=None, sample_weight=None, digits=4, format=None):
        """Receives the optional class names and arguments passed to the report generator function.

        Args:
            class_names: holds the list of class label names provided by the dataset parser. If it is not
                provided when the constructor is called, it will be set by the trainer at runtime.
            sample_weight: sample weights, forwarded to ``sklearn.metrics.classification_report``.
            digits: metrics output digit count, forwarded to ``sklearn.metrics.classification_report``.
            format: output format of the produced log.
        """
        self.class_names = None
        self.sample_weight = sample_weight
        self.digits = digits
        self.pred = None
        self.target = None
        ClassNamesHandler.__init__(self, class_names)
        FormatHandler.__init__(self, format)

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this consumer."""
        return f"{self.__class__.__module__}.{self.__class__.__qualname__}" + \
               f"(class_names={repr(self.class_names)}, sample_weight={repr(self.sample_weight)}, " + \
               f"digits={repr(self.digits)})"

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert isinstance(task, thelper.tasks.Classification), "classif report only impl for classif tasks"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to update function"
        if self.pred is None or self.pred.size != max_iters:
            self.pred = np.asarray([None] * max_iters)
            self.target = np.asarray([None] * max_iters)
        if task.class_names != self.class_names:
            self.class_names = task.class_names
        if target is None or target.numel() == 0:
            # only accumulate results when groundtruth is available
            self.pred[iter_idx] = None
            self.target[iter_idx] = None
            return
        assert pred.dim() == 2 or target.dim() == 1, "current classif report impl only supports batched 1D outputs"
        assert pred.shape[0] == target.shape[0], "prediction/gt tensors batch size mismatch"
        assert pred.shape[1] == len(self.class_names), "unexpected prediction class dimension size"
        self.pred[iter_idx] = pred.topk(1, dim=1)[1].view(pred.shape[0]).tolist()
        self.target[iter_idx] = target.view(target.shape[0]).tolist()

    def gen_report(self, as_dict=False):
        # type: (bool) -> Union[AnyStr, thelper.typedefs.JSON]
        if self.pred is None or self.target is None:
            return None
        pred, target = zip(*[(pred, target) for preds, targets in zip(self.pred, self.target)
                             if targets is not None for pred, target in zip(preds, targets)])
        y_true = np.asarray(target)
        y_pred = np.asarray(pred)
        _y_true = [self.class_names[classid] for classid in y_true]
        _y_pred = [self.class_names[classid] if (0 <= classid < len(self.class_names)) else "<unset>"
                   for classid in y_pred]
        return sklearn.metrics.classification_report(_y_true, _y_pred, sample_weight=self.sample_weight,
                                                     digits=self.digits, output_dict=as_dict)

    def report_text(self):
        # type: () -> Optional[AnyStr]
        """Returns the classification report as a multi-line print-friendly string."""
        report = self.gen_report(as_dict=False)
        return f"\n{report}" if isinstance(report, str) else None

    def report_json(self):
        # type: () -> Optional[AnyStr]
        """Returns the classification report as a JSON formatted string."""
        report = self.gen_report(as_dict=True)
        return json.dumps(report, indent=4) if isinstance(report, str) else None

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying queues."""
        self.pred = None
        self.target = None


@thelper.concepts.detection
class DetectLogger(PredictionConsumer, ClassNamesHandler, FormatHandler):
    """Detection output logger.

    This class provides a simple logging interface for accumulating and saving the bounding boxes of an
    object detector. By default, all detections will be logged. However, a confidence threshold can be set
    to focus on strong predictions if necessary.

    .. todo::
        It also optionally offers tensorboardX-compatible output images that can be saved
        locally or posted to tensorboard for browser-based visualization.

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example consumer; it is used for lookup/printing only
            "logger": {
                # this type is used to instantiate the confusion matrix report object
                "type": "thelper.train.utils.DetectLogger",
                "params": {
                    # (optional) log the three 'best' detections for each target
                    "top_k": 3
                }
            },
            # ...
        }
        # ...

    Attributes:
        top_k: number of 'best' detections to keep for each target bbox (along with the target label).
            If omitted, lists all bounding box predictions by the model after applying IoU and confidence thresholds.
        conf_threshold: threshold used to eliminate uncertain predictions (if they support confidence).
            If confidence is not supported by the model bbox predictions, this parameter is ignored.
        iou_threshold: threshold used to eliminate predictions too far from target (regardless of confidence).
            If omitted, will ignore only completely non-overlapping predicted bounding boxes (:math:`IoU=0`).
            If no target bounding boxes are provided (prediction-only), this parameter is ignored.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        target_name: name of the targeted label (may be 'None' if all classes are used).
        target_idx: index of the targeted label (may be 'None' if all classes are used).
        viz_count: number of tensorboardX images to generate and update at each epoch.
        report_count: number of samples to print in reports (use 'None' if all samples must be printed).
        log_keys: list of metadata field keys to copy from samples into the log for each prediction.
        bbox: array used to store prediction bounding boxes for logging.
        true: array used to store groundtruth labels for logging.
        meta: array used to store metadata pulled from samples for logging.
        format: output format of the produced log (supports: text, CSV, JSON)
    """

    def __init__(self,
                 top_k=None,            # type: Optional[int]
                 conf_threshold=None,   # type: Optional[thelper.typedefs.Number]
                 iou_threshold=None,    # type: Optional[thelper.typedefs.Number]
                 class_names=None,      # type: Optional[List[AnyStr]]
                 target_name=None,      # type: Optional[AnyStr]
                 viz_count=0,           # type: int
                 report_count=None,     # type: Optional[int]
                 log_keys=None,         # type: Optional[List[AnyStr]]
                 format=None,           # type: Optional[AnyStr]
                 ):                     # type: (...) -> None
        """Receives the logging parameters & the optional class label names used to decorate the log."""
        assert top_k is None or isinstance(top_k, int) and top_k > 0, "invalid top-k value"
        assert conf_threshold is None or (isinstance(conf_threshold, (float, int)) and 0 <= conf_threshold <= 1), \
            "detection confidence threshold should be 'None' or number in [0, 1]"
        assert iou_threshold is None or (isinstance(iou_threshold, (int, float)) and 0 <= iou_threshold <= 1), \
            "detection IoU threshold should be 'None' or number in [0, 1]"
        assert isinstance(viz_count, int) and viz_count >= 0, "invalid image count to visualize"
        assert report_count is None or (
            isinstance(report_count, int) and report_count >= 0), "invalid report sample count"
        assert log_keys is None or isinstance(log_keys, list), "invalid list of sample keys to log"
        self.top_k = top_k
        self.target_name = target_name
        self.target_idx = None
        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.viz_count = viz_count
        self.report_count = report_count
        self.log_keys = log_keys if log_keys is not None else []
        self.bbox = None    # type: Optional[thelper.typedefs.DetectionPredictionType]
        self.true = None    # type: Optional[thelper.typedefs.DetectionTargetType]
        self.meta = None    # type: Optional[Dict[AnyStr, List[Any]]]
        ClassNamesHandler.__init__(self, class_names)
        FormatHandler.__init__(self, format)

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this consumer."""
        return f"{self.__class__.__module__}.{self.__class__.__qualname__}" + \
               f"(top_k={repr(self.top_k)}, conf_threshold={repr(self.conf_threshold)}, " + \
               f"class_names={repr(self.class_names)}, target_name={repr(self.target_name)}, " + \
               f"viz_count={repr(self.viz_count)}, report_count={repr(self.report_count)}, " + \
               f"log_keys={repr(self.log_keys)})"

    @ClassNamesHandler.class_names.setter
    def class_names(self, class_names):
        """Sets the class label names that must be predicted by the model."""
        ClassNamesHandler.class_names.fset(self, class_names)
        if self.target_name is not None and self.class_names is not None:
            assert self.target_name in self.class_indices, \
                f"could not find target name {repr(self.target_name)} in class names list"
            self.target_idx = self.class_indices[self.target_name]
        else:
            self.target_idx = None

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.DetectionPredictionType
               target,      # type: thelper.typedefs.DetectionTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert isinstance(task, thelper.tasks.Detection), "detect report only impl for detection tasks"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to update function"
        if self.bbox is None or self.bbox.size != max_iters:
            self.bbox = np.asarray([None] * max_iters)
            self.true = np.asarray([None] * max_iters)
            self.meta = {key: np.asarray([None] * max_iters) for key in self.log_keys}
        if task.class_names != self.class_names:
            self.class_names = task.class_names
        if target is None or len(target) == 0 or all(len(t) == 0 for t in target):
            target = [None] * len(pred)   # simplify unpacking during report generation
        else:
            assert len(pred) == len(target), "prediction/target bounding boxes list batch size mismatch"
            for gt in target:
                assert all(isinstance(bbox, BoundingBox) for bbox in gt), \
                    "detect logger only supports 2D lists of bounding box targets"
        for det in pred:
            assert all(isinstance(bbox, BoundingBox) for bbox in det), \
                "detect logger only supports 2D lists of bounding box predictions"
        self.bbox[iter_idx] = pred
        self.true[iter_idx] = target
        for meta_key in self.log_keys:
            assert meta_key in sample, f"could not extract sample field with key {repr(meta_key)}"
            val = sample[meta_key]
            assert isinstance(val, (list, np.ndarray, torch.Tensor)), f"field {repr(meta_key)} should be batched"
            self.meta[meta_key][iter_idx] = val if isinstance(val, list) else val.tolist()

    def render(self):
        """Returns an image of predicted outputs as a numpy-compatible RGBA image drawn by pyplot."""
        if self.viz_count == 0:
            return None
        if self.bbox is None or self.true is None:
            return None
        raise NotImplementedError  # TODO

    def group_bbox(self,
                   target_bboxes,   # type: List[Optional[BoundingBox]]
                   detect_bboxes,   # type: List[BoundingBox]
                   ):               # type: (...) -> List[Dict[AnyStr, Union[BoundingBox, float, None]]]
        """Groups a sample's detected bounding boxes with target bounding boxes according to configuration parameters.

        Returns a list of detections grouped by target(s) with following format::

            [
                {
                    "target": <associated-target-bbox>,
                    "detect": [
                        {
                            "bbox": <detection-bbox>,
                            "iou": <IoU(detect-bbox, target-bbox)>
                        },
                        ...
                    ]
                },
                ...
            ]

        The associated target bounding box and :math:`IoU` can be ``None`` if no target was provided
        (ie: during inference). In this case, the returned list will have only a single element with all detections
        associated to it. A single element list can also be returned if only one target was specified for this sample.

        When multiple ground truth targets were provided, the returned list will have the same length and ordering as
        these targets. The associated detected bounding boxes will depend on IoU between target/detection combinations.

        All filtering thresholds specified as configuration parameter will be applied for the returned list. Detected
        bounding boxes will also be sorted by highest confidence (if available) or by highest IoU as fallback.
        """
        # remove low confidence and sort by highest
        group_bboxes = sorted(
            [bbox for bbox in (detect_bboxes if detect_bboxes else []) if isinstance(bbox, BoundingBox)],
            key=lambda b: b.confidence if b.confidence is not None else 0, reverse=True)
        if self.conf_threshold:
            group_bboxes = [b for b in group_bboxes if b.confidence is not None and b.confidence >= self.conf_threshold]
        sort_by_iou = all(bbox.confidence is None for bbox in group_bboxes)
        # group according to target count
        target_count = len(target_bboxes)
        if target_count == 0:
            group_bboxes = [{"target": None, "detect": [{"bbox": bbox, "iou": None} for bbox in group_bboxes]}]
        elif target_count == 1:
            sorted_detect = [{"bbox": bbox, "iou": compute_iou(bbox, target_bboxes[0])} for bbox in group_bboxes]
            if sort_by_iou:
                sorted_detect = list(sorted(sorted_detect, key=lambda d: d["iou"], reverse=True))
            group_bboxes = [{"target": target_bboxes[0], "detect": sorted_detect}]
        else:
            # regroup by highest IoU
            target_detects = [[] for _ in range(target_count)]
            for det in group_bboxes:
                # FIXME:
                #  should we do something different if all IoU = 0 (ie: false positive detection)
                #  for now, they will all be stored in the first target, but can be tracked with IoU = 0
                det_target_iou = [compute_iou(det, t) for t in target_bboxes]
                best_iou_idx = int(np.argmax(det_target_iou))
                target_detects[best_iou_idx].append({"bbox": det, "iou": det_target_iou[best_iou_idx]})
            group_bboxes = [{"target": target_bboxes[i],
                             "detect": list(sorted(
                                 target_detects[i], key=lambda d: d["iou"], reverse=True))
                                 if sort_by_iou else target_detects[i]
                            } for i in range(target_count)]     # noqa: PEP8
        # apply filters on grouped results
        if self.iou_threshold:
            for grp in group_bboxes:
                grp["detect"] = [d for d in grp["detect"] if d["iou"] is None or d["iou"] >= self.iou_threshold]
        if self.top_k:
            for grp in group_bboxes:
                grp["detect"] = grp["detect"][:self.top_k]
        return list(group_bboxes)

    def gen_report(self):
        # type: () -> Optional[List[Dict[AnyStr, Any]]]
        """Returns the logged metadata of predicted bounding boxes per sample target in a JSON-like structure.

        For every target bounding box, the corresponding *best*-sorted detections are returned.
        Sample metadata is appended to every corresponding sub-target if any where requested.

        If ``report_count`` was specified, the returned report will be limited to that requested amount of targets.

        .. seealso::
            | :meth:`DetectLogger.group_bbox` for formatting, sorting and filtering details.
        """
        if isinstance(self.report_count, int) and self.report_count <= 0:
            return None
        if self.bbox is None or self.true is None:
            return None
        # flatten batches/samples
        pack = list(zip(*[(*pack,)
                          for packs in zip(self.bbox, self.true, *self.meta.values())
                          for pack in zip(*packs)]))
        data = {key: val for key, val in zip(["detect", "target", *self.meta.keys()], pack)}
        assert all([len(val) == len(data["target"]) for val in data.values()]), "messed up unpacking"

        # flatten targets per batches/samples
        all_targets = []
        sample_count = len(data["target"])
        for sample_idx in range(sample_count):
            if isinstance(self.report_count, int) and self.report_count >= len(all_targets):
                logger.warning(f"report max count {self.report_count} reached at {len(all_targets)} targets "
                               f"(sample {sample_idx}/{sample_count} processed)")
                break
            sample_targets = data["target"][sample_idx]
            sample_detects = data["detect"][sample_idx]
            sample_report = self.group_bbox(sample_targets, sample_detects)
            for target in sample_report:
                for k in self.meta:
                    target[k] = self.meta[k][sample_idx]
                target["target"] = {
                    "bbox": target["target"],
                    "class_name": self.class_names[target["target"].class_id]
                }
            all_targets.extend(sample_report)
        # format everything nicely as json
        for item in all_targets:
            if isinstance(item["target"]["bbox"], BoundingBox):
                item["target"].update(item["target"].pop("bbox").json())
                item["target"]["class_name"] = self.class_names[item["target"]["class_id"]]
            for det in item["detect"]:
                if isinstance(det["bbox"], BoundingBox):
                    det.update(det.pop("bbox").json())
                    det["class_name"] = self.class_names[det["class_id"]]
        return all_targets

    def report_json(self):
        # type: () -> Optional[AnyStr]
        """Returns the logged metadata of predicted bounding boxes as a JSON formatted string."""
        report = self.gen_report()
        if not report:
            return None
        return json.dumps(report, indent=4)

    def report_text(self):
        # type: () -> Optional[AnyStr]
        return self.report_csv()

    def report_csv(self):
        # type: () -> Optional[AnyStr]
        r"""Returns the logged metadata of predicted bounding boxes.

        The returned object is a print-friendly CSV string.

        Note that this string might be very long if the dataset is large or if the model tends to generate a lot of
        detections. The string will contain at least :math:`N_sample \cdot N_target` lines and each line will have
        up to :math:`N_bbox` detections, unless limited by configuration parameters.
        """
        report = self.gen_report()
        if not report:
            return None

        none_str = "unknown"

        def patch_none(to_patch, number_format='2.4f'):  # type: (Any, str) -> str
            if to_patch is None:
                return none_str
            if isinstance(to_patch, float):
                s = f"{{:{number_format}}}"
                return s.format(to_patch)
            return str(to_patch)

        header = "sample,target_name,target_bbox"
        for meta_key in self.log_keys:
            header += f",{str(meta_key)}"
        if self.top_k:
            for k in range(self.top_k):
                header += f",detect_{k + 1}_name,detect_{k + 1}_bbox,detect_{k + 1}_conf,detect_{k + 1}_iou"
        else:
            # unknown count total detections (can be variable)
            header += f",detect_name[N],detect_bbox[N],detect_conf[N],detect_iou[N],(...)[N]"
        lines = [""] * len(report)
        for i, result in enumerate(report):
            target = result["target"]
            detect = result["detect"]
            if not target:
                entry = f"{none_str},{none_str},{none_str}"
            else:
                entry = f"{target['image_id']},{patch_none(target['class_name'])},{patch_none(target['bbox'])}"
            for meta_key in self.log_keys:
                entry += f",{str(target[meta_key])}"
            for det in detect:
                entry += f",{det['class_name']},{det['bbox']},{patch_none(det['confidence'])},{patch_none(det['iou'])}"
            lines[i] = entry
        return "\n".join([header, *lines])

    def reset(self):
        """Toggles a reset of the internal state, emptying storage arrays."""
        self.bbox = None
        self.true = None
        self.meta = None


@thelper.concepts.classification
@thelper.concepts.segmentation
class ConfusionMatrix(PredictionConsumer, ClassNamesHandler):
    """Confusion matrix report interface.

    This class provides a simple interface to ``sklearn.metrics.confusion_matrix`` so that a full
    confusion matrix can be easily reported under a string-based representation. It also offers a
    tensorboardX-compatible output image that can be saved locally or posted to tensorboard for
    browser-based visualization.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example consumer; it is used for lookup/printing only
            "confmat": {
                # this type is used to instantiate the confusion matrix report object
                "type": "thelper.train.utils.ConfusionMatrix",
                # we do not need to provide any parameters to the constructor, defaults are fine
                "params": {
                    # optional parameter that will indicate output as JSON is desired, plain 'text' otherwise
                    "format": "json"
                }
            },
            # ...
        }
        # ...

    Attributes:
        matrix: report generator function, called at evaluation time to generate the output string.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        draw_normalized: defines whether rendered confusion matrices should be normalized or not.
        pred: queue used to store the top-1 (best) predicted class indices at each iteration.
    """

    def __init__(self, class_names=None, draw_normalized=True):
        """Receives the optional class label names used to decorate the output string.

        Args:
            class_names: holds the list of class label names provided by the dataset parser. If it is not
                provided when the constructor is called, it will be set by the trainer at runtime.
            draw_normalized: defines whether rendered confusion matrices should be normalized or not.
        """

        def gen_matrix(y_true, y_pred, _class_names):
            _y_true = [_class_names[classid] for classid in y_true]
            _y_pred = [_class_names[classid] if (0 <= classid < len(_class_names)) else "<unset>" for classid in y_pred]
            return sklearn.metrics.confusion_matrix(_y_true, _y_pred, labels=_class_names)

        self.matrix = gen_matrix
        self.draw_normalized = draw_normalized
        self.pred = None
        self.target = None
        ClassNamesHandler.__init__(self, class_names)

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this consumer."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(class_names={repr(self.class_names)}, draw_normalized={repr(self.draw_normalized)})"

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert isinstance(task, thelper.tasks.Classification), "confmat only impl for classif tasks"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to update function"
        if self.pred is None or self.pred.size != max_iters:
            self.pred = np.asarray([None] * max_iters)
            self.target = np.asarray([None] * max_iters)
        if task.class_names != self.class_names:
            self.class_names = task.class_names
        if target is None or target.numel() == 0:
            # only accumulate results when groundtruth is available
            self.pred[iter_idx] = None
            self.target[iter_idx] = None
            return
        assert pred.dim() == 2 or target.dim() == 1, "current confmat impl only supports batched 1D outputs"
        assert pred.shape[0] == target.shape[0], "prediction/gt tensors batch size mismatch"
        assert pred.shape[1] == len(self.class_names), "unexpected prediction class dimension size"
        self.pred[iter_idx] = pred.topk(1, dim=1)[1].view(pred.shape[0]).tolist()
        self.target[iter_idx] = target.view(target.shape[0]).tolist()

    def report(self):
        """Returns the confusion matrix as a multi-line print-friendly string."""
        if self.pred is None or self.target is None:
            return None
        pred, target = zip(*[(pred, target) for preds, targets in zip(self.pred, self.target)
                             if targets is not None for pred, target in zip(preds, targets)])
        confmat = self.matrix(np.asarray(target), np.asarray(pred), self.class_names)
        return "\n" + thelper.utils.stringify_confmat(confmat, self.class_names)

    def render(self):
        """Returns the confusion matrix as a numpy-compatible RGBA image drawn by pyplot."""
        if self.pred is None or self.target is None:
            return None
        pred, target = zip(*[(pred, target) for preds, targets in zip(self.pred, self.target)
                             if targets is not None for pred, target in zip(preds, targets)])
        confmat = self.matrix(np.asarray(target), np.asarray(pred), self.class_names)
        try:
            fig, ax = thelper.draw.draw_confmat(confmat, self.class_names, normalize=self.draw_normalized)
            array = thelper.draw.fig2array(fig)
            return array
        except AttributeError as e:
            logger.warning(f"failed to render confusion matrix; caught exception:\n{str(e)}")
            # return None if rendering fails (probably due to matplotlib on display-less server)
            return None

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying queues."""
        self.pred = None
        self.target = None


def create_consumers(config):
    """Instantiates and returns the prediction consumers defined in the configuration dictionary.

    All arguments are expected to be handed in through the configuration via a dictionary named 'params'.
    """
    assert isinstance(config, dict), "config should be provided as a dictionary"
    consumers = {}
    for name, consumer_config in config.items():
        assert isinstance(consumer_config, dict), "consumer config should be provided as a dictionary"
        assert "type" in consumer_config and consumer_config["type"], "consumer config missing 'type' field"
        consumer_type = thelper.utils.import_class(consumer_config["type"])
        consumer_params = thelper.utils.get_key_def(["params", "parameters"], consumer_config, {})
        try:
            consumer = consumer_type(**consumer_params)
        except Exception:
            logger.error(f"failed to create consumer {consumer_config['type']} with params:\n\t{str(consumer_params)}")
            raise
        assert isinstance(consumer, PredictionConsumer), \
            "invalid consumer type, must derive from PredictionConsumer interface"
        consumers[name] = consumer
    return consumers


def create_trainer(session_name,    # type: AnyStr
                   save_dir,        # type: AnyStr
                   config,          # type: thelper.typedefs.ConfigDict
                   model,           # type: thelper.typedefs.ModelType
                   task,            # type: thelper.tasks.Task
                   loaders,         # type: thelper.typedefs.MultiLoaderType
                   ckptdata=None    # type: Optional[thelper.typedefs.CheckpointContentType]
                   ):               # type: (...) -> thelper.train.Trainer
    """Instantiates the trainer object based on the type contained in the config dictionary.

    The trainer type is expected to be in the configuration dictionary's `trainer` field, under the `type` key. For more
    information on the configuration, refer to :class:`thelper.train.base.Trainer`. The instantiated type must be
    compatible with the constructor signature of :class:`thelper.train.base.Trainer`. The object's constructor will
    be given the full config dictionary and the checkpoint data for resuming the session (if available).

    If the trainer type is missing, it will be automatically deduced based on the task object.

    Args:
        session_name: name of the training session used for printing and to create internal tensorboardX directories.
        save_dir: path to the session directory where logs and checkpoints will be saved.
        config: full configuration dictionary that will be parsed for trainer parameters and saved in checkpoints.
        model: model to train/evaluate; should be compatible with :class:`thelper.nn.utils.Module`.
        task: global task interface defining the type of model and training goal for the session.
        loaders: a tuple containing the training/validation/test data loaders (a loader can be ``None`` if empty).
        ckptdata: raw checkpoint to parse data from when resuming a session (if ``None``, will start from scratch).

    Returns:
        The fully-constructed trainer object, ready to begin model training/evaluation.

    .. seealso::
        | :class:`thelper.train.base.Trainer`

    """
    assert "trainer" in config and config["trainer"], "session configuration dictionary missing 'trainer' section"
    trainer_config = config["trainer"]
    if "type" not in trainer_config:
        if isinstance(task, thelper.tasks.Classification):
            trainer_type = thelper.train.ImageClassifTrainer
        elif isinstance(task, thelper.tasks.Detection):
            trainer_type = thelper.train.ObjDetectTrainer
        elif isinstance(task, thelper.tasks.Regression):
            trainer_type = thelper.train.RegressionTrainer
        elif isinstance(task, thelper.tasks.Segmentation):
            trainer_type = thelper.train.ImageSegmTrainer
        else:
            raise AssertionError(f"unknown trainer type required for task '{str(task)}'")
    else:
        trainer_type = thelper.utils.import_class(trainer_config["type"])
    return trainer_type(session_name, save_dir, model, task, loaders, config, ckptdata=ckptdata)


# noinspection PyUnusedLocal
def _draw_wrapper(task,         # type: thelper.tasks.utils.Task
                  input,        # type: thelper.typedefs.InputType
                  pred,         # type: thelper.typedefs.AnyPredictionType
                  target,       # type: thelper.typedefs.AnyTargetType
                  sample,       # type: thelper.typedefs.SampleType
                  loss,         # type: Optional[float]
                  iter_idx,     # type: int
                  max_iters,    # type: int
                  epoch_idx,    # type: int
                  max_epochs,   # type: int
                  # extra params added by callback interface below
                  output_path,  # type: str
                  save,         # type: bool
                  # all extra params will be forwarded to the display call
                  **kwargs,     # type: Any
                  ):            # type: (...) -> None
    """Wrapper to :func:`thelper.draw.draw` used as a callback entrypoint for trainers."""
    res = thelper.draw.draw(task=task, input=input, pred=pred, target=target, **kwargs)
    if save:
        assert isinstance(res, tuple) and len(res) == 2, "unexpected redraw output (should be 2-elem tuple)"
        if isinstance(res[1], np.ndarray):
            assert "path" in sample and isinstance(sample["path"], list) and len(sample["path"]) == 1, \
                "unexpected draw format (current implementation needs batch size = 1, and path metadata)"
            os.makedirs(output_path, exist_ok=True)
            filepath = os.path.join(output_path, os.path.basename(sample["path"][0]))
            cv.imwrite(filepath, res[1])
        else:
            # we're displaying with matplotlib, and have no clue on how to save the output
            raise NotImplementedError



import cv2 as cv
import numpy as np
import PIL.Image
import pytest
import torch

import thelper


def test_notransform():
    op = thelper.transforms.NoTransform()
    sample = np.random.randn(5, 5, 5)
    transf = op(sample)
    assert transf is sample
    transf_inv = op.invert(transf)
    assert transf_inv is sample
    op2 = eval(repr(op))
    assert type(op) == type(op2)
    assert op.__dict__ == op2.__dict__


def test_tonumpy():
    op = thelper.transforms.ToNumpy()
    sample = np.random.randn(5, 5, 5)
    cvt_sample = op(sample)
    assert cvt_sample is sample
    sample = torch.from_numpy(np.random.randint(0, 255, (32, 32, 3), dtype=np.uint8))
    cvt_sample = op(sample)
    assert isinstance(cvt_sample, np.ndarray)
    assert np.array_equal(cvt_sample, sample.numpy())
    orig_sample = np.random.randint(0, 255, (32, 32, 3), dtype=np.uint8)
    sample = PIL.Image.fromarray(orig_sample)
    cvt_sample = op(sample)
    assert isinstance(cvt_sample, np.ndarray)
    assert np.array_equal(cvt_sample, orig_sample)
    op = thelper.transforms.ToNumpy(reorder_bgr=True)
    cvt_sample = op(sample)
    assert isinstance(cvt_sample, np.ndarray)
    assert np.array_equal(cvt_sample, orig_sample[..., ::-1])
    with pytest.raises(RuntimeError):
        _ = op.invert(cvt_sample)
    op2 = eval(repr(op))
    assert type(op) == type(op2)
    assert op.__dict__ == op2.__dict__


def test_centercrop():
    with pytest.raises((AssertionError, ValueError)):
        _ = thelper.transforms.CenterCrop(size=-1, borderval=99)
    with pytest.raises((AssertionError, ValueError)):
        _ = thelper.transforms.CenterCrop(size=(10, 10, 10), borderval=99)
    op1 = thelper.transforms.CenterCrop(size=10, borderval=99)
    op2 = thelper.transforms.CenterCrop(size=(10, 10), borderval=99)
    assert op1.size == op2.size
    array = []
    idx = 0
    for i in range(12):
        array.append([])
        for j in range(12):
            array[i].append([i, j, idx])
            idx += 1
    sample = np.asarray(array)
    out1 = op1(sample)
    out2 = op2(sample)
    assert np.array_equal(out1, out2)
    assert out1.shape == (10, 10, 3)
    assert np.array_equal(out1[0, :, 0], [1] * 10)
    assert np.array_equal(out1[-1, :, 0], [10] * 10)
    assert np.array_equal(out1[:, 0, 1], [1] * 10)
    assert np.array_equal(out1[:, -1, 1], [10] * 10)
    assert out1[-1, -1, 2] == 130
    with pytest.raises((AssertionError, RuntimeError)):
        _ = op1.invert(out1)
    op3 = eval(repr(op1))
    assert type(op1) == type(op3)
    assert op1.__dict__ == op3.__dict__


def test_randomresizedcrop():
    with pytest.raises((AssertionError, ValueError)):
        _ = thelper.transforms.RandomResizedCrop(output_size=())
    op1 = thelper.transforms.RandomResizedCrop(output_size=10)
    op2 = thelper.transforms.RandomResizedCrop(output_size=(10, 10))
    assert op1.output_size == op2.output_size
    with pytest.raises((AssertionError, TypeError)):
        _ = thelper.transforms.RandomResizedCrop(output_size="something")
    op3 = thelper.transforms.RandomResizedCrop(output_size=None, input_size=(20, 30), ratio=(0.5, 1.0))
    op4 = thelper.transforms.RandomResizedCrop(output_size=None, input_size=(30, 20), ratio=0.66)
    assert op3.output_size == op4.output_size and op3.input_size == op4.input_size
    op5 = thelper.transforms.RandomResizedCrop(output_size=None, input_size=((30, 30), (20, 20)), ratio=None)
    with pytest.raises((AssertionError, ValueError)):
        _ = thelper.transforms.RandomResizedCrop(output_size=None, input_size=((20, 30), (20, 30)), ratio=0.66)
    op6 = thelper.transforms.RandomResizedCrop(output_size=None, input_size=((20, 30), (30, 20)), ratio=None)
    assert op5.output_size == op6.output_size and op5.input_size == op6.input_size
    with pytest.raises((AssertionError, TypeError)):
        _ = thelper.transforms.RandomResizedCrop(output_size=None, input_size="something")
    with pytest.raises((AssertionError, ValueError)):
        _ = thelper.transforms.RandomResizedCrop(output_size=None, input_size=(0.0, 1.0))
    op7 = thelper.transforms.RandomResizedCrop(output_size=None, input_size=(0.1, 1.0), probability=0.5)
    assert op7.probability == 0.5
    with pytest.raises((AssertionError, ValueError)):
        _ = thelper.transforms.RandomResizedCrop(output_size=None, input_size=(0.1, 1.0), probability=-1)
    op8 = thelper.transforms.RandomResizedCrop(output_size=None, flags="cv2.INTER_LINEAR")
    assert op8.flags == cv.INTER_LINEAR



import copy
import os
import shutil

import numpy as np
import pytest
import torch

import thelper

test_save_path = ".pytest_cache"

test_create_simple_path = os.path.join(test_save_path, "simple")
test_create_simple_images_path = os.path.join(test_save_path, "simple_images")


@pytest.fixture
def simple_config(request):
    def fin():
        shutil.rmtree(test_create_simple_path, ignore_errors=True)
        shutil.rmtree(test_create_simple_images_path, ignore_errors=True)
    fin()
    request.addfinalizer(fin)
    os.makedirs(test_create_simple_images_path, exist_ok=True)
    for cls in range(10):
        os.makedirs(os.path.join(test_create_simple_images_path, str(cls)), exist_ok=True)
        for idx in range(10):
            open(os.path.join(test_create_simple_images_path, str(cls), str(idx) + ".jpg"), "a").close()
    return {
        "name": "simple",
        "bypass_queries": True,
        "datasets": {
            "dset": {
                "type": "thelper.data.ImageFolderDataset",
                "params": {
                    "root": test_create_simple_images_path
                }
            }
        },
        "loaders": {
            "shuffle": True,
            "workers": 0,
            "batch_size": 32,
            "skip_class_balancing": True,
            "train_split": {
                "dset": 0.9
            },
            "valid_split": {
                "dset": 0.1
            }
        },
        "model": {
            "type": "thelper.nn.resnet.ResNet"
        },
        "trainer": {
            "type": "thelper.train.ImageClassifTrainer",
            "epochs": 2,
            "optimization": {
                "loss": {
                    "type": "torch.nn.CrossEntropyLoss"
                },
                "optimizer": {
                    "type": "torch.optim.Adam",
                    "params": {
                        "lr": 0.001
                    }
                }
            }
        }
    }


def test_save_load_simple_config(simple_config):
    orig_config = copy.deepcopy(simple_config)
    assert thelper.utils.get_config_session_name(orig_config) == "simple"
    json_config_path = os.path.join(test_save_path, "dummy_name.json")
    thelper.utils.save_config(simple_config, json_config_path)
    new_config = thelper.utils.load_config(json_config_path)
    assert new_config == orig_config
    yaml_config_path = os.path.join(test_save_path, "dummy_name.yaml")
    thelper.utils.save_config(simple_config, yaml_config_path)
    new_config = thelper.utils.load_config(yaml_config_path)
    assert new_config == orig_config
    pkl_config_path = os.path.join(test_save_path, "dummy_name.pkl")
    thelper.utils.save_config(simple_config, pkl_config_path)
    new_config = thelper.utils.load_config(pkl_config_path)
    assert new_config == orig_config
    txt_config_path = os.path.join(test_save_path, "dummy_name.txt")
    thelper.utils.save_config(simple_config, txt_config_path, as_json=True)
    new_config = thelper.utils.load_config(txt_config_path, as_json=True)
    assert new_config == orig_config


def test_create_session_nameless(simple_config, mocker):
    fake_train = mocker.patch.object(thelper.train.base.Trainer, "train")
    fake_eval = mocker.patch.object(thelper.train.base.Trainer, "eval")
    del simple_config["name"]
    with pytest.raises(AssertionError):
        thelper.cli.create_session(simple_config, test_save_path)
    assert fake_train.call_count == 0
    assert fake_eval.call_count == 0
    config_save_path = os.path.join(test_save_path, "dummy.json")
    thelper.utils.save_config(simple_config, config_save_path)
    new_config = thelper.utils.load_config(config_save_path, add_name_if_missing=False)
    assert thelper.utils.get_config_session_name(new_config) is None
    assert new_config == simple_config
    new_config = thelper.utils.load_config(config_save_path)
    assert thelper.utils.get_config_session_name(new_config) == "dummy"


def test_create_session_train(simple_config, mocker):
    fake_train = mocker.patch.object(thelper.train.base.Trainer, "train")
    fake_eval = mocker.patch.object(thelper.train.base.Trainer, "eval")
    thelper.cli.create_session(simple_config, test_save_path)
    assert fake_train.call_count == 1
    assert fake_eval.call_count == 0
    assert os.path.isdir(test_create_simple_path)


def test_create_session_eval(simple_config, mocker):
    fake_train = mocker.patch.object(thelper.train.base.Trainer, "train")
    fake_eval = mocker.patch.object(thelper.train.base.Trainer, "eval")
    del simple_config["loaders"]["train_split"]
    thelper.cli.create_session(simple_config, test_save_path)
    assert fake_train.call_count == 0
    assert fake_eval.call_count == 1
    assert os.path.isdir(test_create_simple_path)


def test_resume_session(simple_config, mocker):
    log_iter = 0

    def get_log_stamp():
        nonlocal log_iter
        log_iter += 1
        return "fake-20000101-%06d" % log_iter

    fake_logstamp = mocker.patch("thelper.utils.get_log_stamp")
    fake_logstamp.side_effect = get_log_stamp
    with pytest.raises(AssertionError):
        thelper.cli.resume_session(None, test_save_path, simple_config)
    with pytest.raises(AssertionError):
        thelper.cli.resume_session({"dummy": None}, test_save_path)
    with pytest.raises(AssertionError):
        thelper.cli.resume_session({"config": {}}, test_save_path)
    fake_train = mocker.patch.object(thelper.train.classif.ImageClassifTrainer, "train_epoch", return_value=[0, 0])
    fake_eval = mocker.patch.object(thelper.train.classif.ImageClassifTrainer, "eval_epoch")
    with pytest.raises((FileNotFoundError, AssertionError)):
        _ = thelper.utils.load_checkpoint(test_create_simple_path)
    thelper.cli.create_session(simple_config, test_save_path)
    assert fake_train.call_count == 2
    assert fake_eval.call_count == 2
    ckptdata = thelper.utils.load_checkpoint(test_create_simple_path)
    ckptdata_fake = {key: val for key, val in ckptdata.items() if key != "task"}
    with pytest.raises(AssertionError):
        thelper.cli.resume_session(ckptdata_fake, test_save_path, eval_only=True)
    nameless_ckptdata = copy.deepcopy(ckptdata)
    del nameless_ckptdata["config"]["name"]
    with pytest.raises(AssertionError):
        thelper.cli.resume_session(nameless_ckptdata, test_save_path, eval_only=True)
    old_task = ckptdata["task"]
    thelper.cli.resume_session(ckptdata, test_save_path, eval_only=True)
    latest_ckptdata = thelper.utils.load_checkpoint(test_create_simple_path, always_load_latest=True)
    assert old_task == latest_ckptdata["task"]
    assert fake_train.call_count == 2
    assert fake_eval.call_count == 3
    thelper.cli.resume_session(ckptdata, test_save_path)
    latest_ckptdata = thelper.utils.load_checkpoint(test_create_simple_path, always_load_latest=True)
    assert old_task == latest_ckptdata["task"]
    assert fake_train.call_count == 3
    assert fake_eval.call_count == 4
    override_config = ckptdata["config"]
    override_config["trainer"]["epochs"] = 3
    thelper.cli.resume_session(ckptdata, test_save_path, override_config)
    latest_ckptdata = thelper.utils.load_checkpoint(test_create_simple_path, always_load_latest=True)
    assert old_task == latest_ckptdata["task"]
    assert fake_train.call_count == 5
    assert fake_eval.call_count == 6
    override_config = ckptdata["config"]
    new_finetune_task = {
        "type": "thelper.tasks.Classification",
        "params": {
            "class_names": [
                "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"
            ],
            "input_key": "image",
            "label_key": "label"
        }
    }
    override_config["datasets"]["dset"]["task"] = new_finetune_task
    fake_query = mocker.patch("thelper.utils.query_string", return_value="new")
    thelper.cli.resume_session(ckptdata, test_save_path)
    latest_ckptdata = thelper.utils.load_checkpoint(test_create_simple_path, always_load_latest=True)
    assert str(thelper.tasks.create_task(new_finetune_task)) == latest_ckptdata["task"]
    assert fake_query.call_count == 1
    _ = mocker.patch("thelper.utils.query_string", return_value="old")
    thelper.cli.resume_session(ckptdata, test_save_path)
    latest_ckptdata = thelper.utils.load_checkpoint(test_create_simple_path, always_load_latest=True)
    assert old_task == latest_ckptdata["task"]


def test_visualize_data(simple_config, mocker):
    fake_draw = mocker.patch("thelper.draw.draw")
    fake_imread = mocker.patch("cv2.imread", return_value=["ohhai"])
    fake_config_bad_viz = copy.deepcopy(simple_config)
    fake_config_bad_viz["viz"] = "bad"
    with pytest.raises(AssertionError):
        thelper.cli.visualize_data(fake_config_bad_viz)
    fake_config_bad_viz["viz"] = {"kwargs": "bad"}
    with pytest.raises(AssertionError):
        thelper.cli.visualize_data(fake_config_bad_viz)
    assert fake_draw.call_count == 0
    config_viz_no_loaders = copy.deepcopy(simple_config)
    del config_viz_no_loaders["loaders"]
    query_idx = 0

    def query_dset(*args, **kwargs):
        nonlocal query_idx
        if query_idx == 0:
            query_idx += 1
            return "dset"
        else:
            return "quit"
    fake_query = mocker.patch("thelper.utils.query_string")
    fake_query.side_effect = query_dset
    thelper.cli.visualize_data(config_viz_no_loaders)
    assert fake_draw.call_count == 100
    assert fake_imread.call_count == 100
    query_idx = 0

    def query_loader(*args, **kwargs):
        nonlocal query_idx
        if query_idx == 0:
            query_idx += 1
            return "train"
        elif query_idx == 1:
            query_idx += 1
            return "valid"
        elif query_idx == 2:
            query_idx += 1
            return "test"
        else:
            return "quit"
    fake_query.side_effect = query_loader
    thelper.cli.visualize_data(simple_config)
    assert fake_draw.call_count == 104
    assert fake_imread.call_count == 200


@pytest.fixture
def annot_config(request):
    def fin():
        shutil.rmtree(test_create_simple_path, ignore_errors=True)
        shutil.rmtree(test_create_simple_images_path, ignore_errors=True)
    fin()
    request.addfinalizer(fin)
    os.makedirs(test_create_simple_images_path, exist_ok=True)
    for cls in range(10):
        os.makedirs(os.path.join(test_create_simple_images_path, str(cls)), exist_ok=True)
        for idx in range(10):
            open(os.path.join(test_create_simple_images_path, str(cls), str(idx) + ".jpg"), "a").close()
    return {
        "name": "simple",
        "bypass_queries": True,
        "datasets": {
            "output": {
                "type": "thelper.data.ImageFolderDataset",
                "params": {
                    "root": test_create_simple_images_path
                }
            }
        },
        "annotator": {
            "type": "thelper.gui.ImageSegmentAnnotator",
            "params": {
                "sample_input_key": "image",
                "labels": [
                    {"id": 255, "name": "foreground", "color": [0, 0, 255]}
                ]
            }
        }
    }


def test_annotate_data(annot_config, mocker):
    import thelper.gui
    fake_run = mocker.patch.object(thelper.gui.ImageSegmentAnnotator, "run")
    fake_imread = mocker.patch("cv2.imread", return_value=np.random.randint(0, 255, (256, 256, 3), dtype=np.uint8))
    _ = mocker.patch("cv2.setMouseCallback")
    _ = mocker.patch("cv2.namedWindow")
    _ = mocker.patch("cv2.waitKey")
    _ = mocker.patch("cv2.imshow")
    _ = mocker.patch("thelper.gui.create_key_listener")
    fake_config = copy.deepcopy(annot_config)
    del fake_config["name"]
    with pytest.raises(AssertionError):
        thelper.cli.annotate_data(fake_config, test_save_path)
    assert fake_run.call_count == 0
    thelper.cli.annotate_data(annot_config, test_save_path)
    assert fake_run.call_count == 1
    assert fake_imread.call_count == 1


@pytest.fixture
def split_config(request):
    def fin():
        shutil.rmtree(test_create_simple_path, ignore_errors=True)
    fin()
    request.addfinalizer(fin)
    os.makedirs(test_create_simple_images_path, exist_ok=True)
    for cls in range(10):
        os.makedirs(os.path.join(test_create_simple_images_path, str(cls)), exist_ok=True)
        for idx in range(10):
            open(os.path.join(test_create_simple_images_path, str(cls), str(idx) + ".jpg"), "a").close()
    return {
        "name": "simple",
        "bypass_queries": True,
        "datasets": {
            "dset": {
                "type": "thelper.data.ImageFolderDataset",
                "params": {
                    "root": test_create_simple_images_path
                }
            }
        },
        "loaders": {
            "shuffle": True,
            "workers": 0,
            "batch_size": 1,
            "skip_class_balancing": True,
            "train_split": {
                "dset": 0.9
            },
            "valid_split": {
                "dset": 0.1
            }
        }
    }


def test_split_data(split_config, mocker):
    fake_config = copy.deepcopy(split_config)
    fake_create = mocker.patch("thelper.data.create_hdf5")
    del fake_config["name"]
    with pytest.raises(AssertionError):
        thelper.cli.split_data(fake_config, test_save_path)
    fake_config = copy.deepcopy(split_config)
    del fake_config["loaders"]
    with pytest.raises(AssertionError):
        thelper.cli.split_data(fake_config, test_save_path)
    fake_config = copy.deepcopy(split_config)
    fake_config["split"] = "dummy"
    with pytest.raises(AssertionError):
        thelper.cli.split_data(fake_config, test_save_path)
    fake_config = copy.deepcopy(split_config)
    fake_config["split"] = {"compression": "dummy"}
    with pytest.raises(AssertionError):
        thelper.cli.split_data(fake_config, test_save_path)
    thelper.cli.split_data(split_config, test_save_path)
    assert fake_create.call_count == 1


@pytest.fixture
def export_config(request):
    def fin():
        shutil.rmtree(test_create_simple_path, ignore_errors=True)
        shutil.rmtree(test_create_simple_images_path, ignore_errors=True)

    fin()
    request.addfinalizer(fin)
    os.makedirs(test_create_simple_images_path, exist_ok=True)
    for cls in range(10):
        os.makedirs(os.path.join(test_create_simple_images_path, str(cls)), exist_ok=True)
        for idx in range(10):
            open(os.path.join(test_create_simple_images_path, str(cls), str(idx) + ".jpg"), "a").close()
    return {
        "name": "simple",
        "bypass_queries": True,
        "model": {
            "type": "torchvision.models.resnet.resnet18",
            "params": {
                "pretrained": True
            }
        },
        "export": {
            "ckpt_name": "test-resnet18-imagenet.pth",
            "trace_name": "test-resnet18-imagenet.zip",
            "save_raw": True,
            "trace_input": "torch.rand(1, 3, 224, 224)",
            "task": {
                "type": "thelper.tasks.Classification",
                "params": {
                    "class_names": "tests/meta/imagenet_classes.json",
                    "input_key": "0",
                    "label_key": "1"
                }
            }
        }
    }


def test_export_model_trace(export_config):
    fake_config = copy.deepcopy(export_config)
    del fake_config["name"]
    with pytest.raises(AssertionError):
        thelper.cli.export_model(fake_config, test_save_path)
    fake_config = copy.deepcopy(export_config)
    fake_config["export"] = "dummy"
    with pytest.raises(AssertionError):
        thelper.cli.export_model(fake_config, test_save_path)
    thelper.cli.export_model(export_config, test_save_path)
    export_ckpt_path = os.path.join(test_create_simple_path, export_config["export"]["ckpt_name"])
    assert os.path.isfile(export_ckpt_path)
    ckptdata = thelper.utils.load_checkpoint(export_ckpt_path)
    model = thelper.nn.create_model(config=None, task=None, ckptdata=ckptdata)
    assert model(eval(export_config["export"]["trace_input"])).shape == (1, 1000)


def test_export_model_weights(export_config):
    export_config = copy.deepcopy(export_config)
    del export_config["export"]["trace_input"]
    del export_config["export"]["trace_name"]
    del export_config["export"]["task"]
    export_config["datasets"] = {
        "dset": {
            "type": "thelper.data.ImageFolderDataset",
            "params": {
                "root": test_create_simple_images_path
            }
        }
    }
    thelper.cli.export_model(export_config, test_save_path)
    export_ckpt_path = os.path.join(test_create_simple_path, export_config["export"]["ckpt_name"])
    assert os.path.isfile(export_ckpt_path)
    ckptdata = thelper.utils.load_checkpoint(export_ckpt_path)
    model = thelper.nn.create_model(config=None, task=None, ckptdata=ckptdata)
    assert model(torch.rand(1, 3, 224, 224)).shape == (1, 10)



"""Dataset loaders module.

This module contains a dataset loader specialization used to properly seed samplers and workers.
"""

import copy
import inspect
import logging
import math
import random
import sys
import time
from collections import Counter

import numpy as np
import torch
import torch.utils.data
import torch.utils.data.sampler
import tqdm

import thelper.tasks
import thelper.transforms
import thelper.utils

logger = logging.getLogger(__name__)


def default_collate(batch, force_tensor=True):
    """Puts each data field into a tensor with outer dimension batch size.

    This function is copied from PyTorch's `torch.utils.data._utils.collate.default_collate`, but
    additionally supports custom objects from the framework (such as bounding boxes). These will not
    be converted to tensors, and it will be up to the trainer to handle them accordingly.

    See ``torch.utils.data.DataLoader`` for more information.
    """
    from torch._six import container_abcs, string_classes, int_classes
    error_msg_fmt = "batch must contain tensors, numbers, dicts or lists; found {}"
    torch_ver = [int(v) for v in torch.__version__.split(".")]
    elem_type = type(batch[0])
    if any([b is None for b in batch]):
        assert all([b is None for b in batch]), "cannot mix ``None`` and non-``None`` types"
        return None  # compress and return entire field as unavailable
    elif isinstance(batch[0], torch.Tensor):
        out = None
        if torch_ver[0] > 1 or torch_ver[1] > 1:  # ver > 1.1
            if torch.utils.data.get_worker_info() is not None:
                numel = sum([x.numel() for x in batch])
                storage = batch[0].storage()._new_shared(numel)
                out = batch[0].new(storage)
            return torch.stack(batch, 0, out=out)
        elif torch_ver[0] == 1 and torch_ver[1] == 1:  # ver == 1.1
            if torch.utils.data._utils.collate._use_shared_memory:
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum([x.numel() for x in batch])
                storage = batch[0].storage()._new_shared(numel)
                out = batch[0].new(storage)
            return torch.stack(batch, 0, out=out)
        else:  # ver < 1.1
            if torch.utils.data.dataloader._use_shared_memory:
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum([x.numel() for x in batch])
                storage = batch[0].storage()._new_shared(numel)
                out = batch[0].new(storage)
            return torch.stack(batch, 0, out=out)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' and \
            elem_type.__name__ != 'string_':
        elem = batch[0]
        if elem_type.__name__ == 'ndarray':
            # array of string classes and object
            if torch_ver[0] > 1 or torch_ver[1] > 0:  # ver > 1.0
                if torch.utils.data._utils.collate.np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                    raise TypeError(error_msg_fmt.format(elem.dtype))
            else:  # ver <= 1.0
                import re
                if re.search('[SaUO]', elem.dtype.str) is not None:
                    raise TypeError(error_msg_fmt.format(elem.dtype))
            return default_collate([torch.from_numpy(b) for b in batch], force_tensor=force_tensor)
        if elem.shape == ():  # scalars
            if torch_ver[0] > 1 or torch_ver[1] > 1:  # ver > 1.1
                py_type = float if elem.dtype.name.startswith('float') else int
                if torch_ver[0] > 1 or torch_ver[1] > 0:  # ver > 1.0
                    return torch.utils.data._utils.collate.numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
                else:  # ver <= 1.0
                    return torch.utils.data.dataloader.numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
            else:  # ver <= 1.1
                return torch.as_tensor(batch)
    elif isinstance(batch[0], float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(batch[0], int_classes):
        return torch.tensor(batch)
    elif isinstance(batch[0], string_classes):
        return batch
    elif isinstance(batch[0], container_abcs.Mapping):
        return {key: default_collate([d[key] for d in batch], force_tensor=force_tensor) for key in batch[0]}
    elif isinstance(batch[0], tuple) and hasattr(batch[0], '_fields'):  # namedtuple
        return type(batch[0])(*(default_collate(samples, force_tensor=force_tensor) for samples in zip(*batch)))
    elif isinstance(batch[0], container_abcs.Sequence):
        if isinstance(batch, list) and all([isinstance(l, list) for l in batch]) and \
                all([isinstance(b, thelper.data.BoundingBox) for l in batch for b in l]):
            return batch
        transposed = zip(*batch)
        return [default_collate(samples, force_tensor=force_tensor) for samples in transposed]
    if force_tensor:
        raise TypeError((error_msg_fmt.format(type(batch[0]))))
    return batch


class DataLoader(torch.utils.data.DataLoader):
    """Specialized data loader used to load minibatches from a dataset parser.

    This specialization handles the seeding of samplers and workers.

    See ``torch.utils.data.DataLoader`` for more information on attributes/methods.
    """
    def __init__(self, *args, seeds=None, epoch=0, collate_fn=default_collate, **kwargs):
        super().__init__(*args, collate_fn=collate_fn, worker_init_fn=self._worker_init_fn, **kwargs)
        self.seeds = {}
        if seeds is not None:
            if not isinstance(seeds, dict):
                raise AssertionError("unexpected seed pack type")
            self.seeds = seeds
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch
        self.num_workers = kwargs["num_workers"] if "num_workers" in kwargs else 0

    def __iter__(self):
        """Advances the epoch number for the workers initialization function."""
        self.set_epoch(self.epoch)  # preset for all attributes
        if self.num_workers == 0:
            if "torch" in self.seeds:
                torch.manual_seed(self.seeds["torch"] + self.epoch)
                torch.cuda.manual_seed_all(self.seeds["torch"] + self.epoch)
            if "numpy" in self.seeds:
                np.random.seed(self.seeds["numpy"] + self.epoch)
            if "random" in self.seeds:
                random.seed(self.seeds["random"] + self.epoch)
        result = super().__iter__()
        self.epoch += 1
        return result

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to offset RNG states for the workers and the sampler."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch
        if self.sampler is not None:
            if hasattr(self.sampler, "set_epoch") and callable(self.sampler.set_epoch):
                self.sampler.set_epoch(self.epoch)
        if hasattr(self, "dataset"):
            if hasattr(self.dataset, "set_epoch") and callable(self.dataset.set_epoch):
                self.dataset.set_epoch(epoch)
            if hasattr(self.dataset, "transforms"):
                if hasattr(self.dataset.transforms, "set_epoch") and callable(self.dataset.transforms.set_epoch):
                    self.dataset.transforms.set_epoch(epoch)

    def _worker_init_fn(self, worker_id):
        """Sets up the RNGs state of each worker based on their unique id and the epoch number."""
        seed_offset = self.num_workers * self.epoch
        if "torch" in self.seeds:
            torch.manual_seed(self.seeds["torch"] + seed_offset + worker_id)
            torch.cuda.manual_seed_all(self.seeds["torch"] + seed_offset + worker_id)
        if "numpy" in self.seeds:
            np.random.seed(self.seeds["numpy"] + seed_offset + worker_id)
        if "random" in self.seeds:
            random.seed(self.seeds["random"] + seed_offset + worker_id)

    @property
    def sample_count(self):
        return len(self.sampler) if self.sampler is not None else len(self.dataset)


class LoaderFactory:
    """Factory used for preparing and splitting dataset parsers into usable data loader objects.

    This class is responsible for parsing the parameters contained in the 'loaders' field of a
    configuration dictionary, instantiating the data loaders, and shuffling/splitting the samples.
    An example configuration is presented in :func:`thelper.data.utils.create_loaders`.

    .. seealso::
        | :func:`thelper.data.utils.create_loaders`
        | :func:`thelper.transforms.utils.load_augments`
        | :func:`thelper.transforms.utils.load_transforms`
    """

    def __init__(self, config):
        """Receives and parses the data configuration dictionary."""
        logger.debug("loading data configuration...")
        default_batch_size = 1
        if "batch_size" in config:
            if any([v in config for v in ["train_batch_size", "valid_batch_size", "test_batch_size"]]):
                raise AssertionError("specifying 'batch_size' overrides all other (loader-specific) values")
            default_batch_size = int(thelper.utils.get_key("batch_size", config))
        self.train_batch_size = thelper.utils.get_key_def("train_batch_size", config, default_batch_size)
        self.valid_batch_size = thelper.utils.get_key_def("valid_batch_size", config, default_batch_size)
        self.test_batch_size = thelper.utils.get_key_def("test_batch_size", config, default_batch_size)
        assert self.train_batch_size > 0, "batch size should be strictly positive integer"
        assert self.valid_batch_size > 0, "batch size should be strictly positive integer"
        assert self.test_batch_size > 0, "batch size should be strictly positive integer"
        self.train_scale = thelper.utils.get_key_def("train_scale", config, 1.0)
        self.valid_scale = thelper.utils.get_key_def("valid_scale", config, 1.0)
        self.test_scale = thelper.utils.get_key_def("test_scale", config, 1.0)
        default_collate_fn = default_collate
        if "collate_fn" in config:
            if any([v in config for v in ["train_collate_fn", "valid_collate_fn", "test_collate_fn"]]):
                raise AssertionError("specifying 'collate_fn' overrides all other (loader-specific) values")
            default_collate_fn = self._get_collate_fn(config["collate_fn"])
        self.train_collate_fn = self._get_collate_fn(thelper.utils.get_key_def("train_collate_fn", config, default_collate_fn))
        self.valid_collate_fn = self._get_collate_fn(thelper.utils.get_key_def("valid_collate_fn", config, default_collate_fn))
        self.test_collate_fn = self._get_collate_fn(thelper.utils.get_key_def("test_collate_fn", config, default_collate_fn))
        self.train_shuffle = thelper.utils.str2bool(thelper.utils.get_key_def(["shuffle", "train_shuffle"], config, True))
        self.valid_shuffle = thelper.utils.str2bool(thelper.utils.get_key_def(["shuffle", "valid_shuffle"], config, False))
        self.test_shuffle = thelper.utils.str2bool(thelper.utils.get_key_def(["shuffle", "test_shuffle"], config, False))
        np.random.seed()  # for seed generation below (if needed); will be reseeded afterwards
        test_seed = self._get_seed(["test_seed", "test_split_seed"], config, (int, str))
        valid_seed = self._get_seed(["valid_seed", "valid_split_seed"], config, (int, str))
        torch_seed = self._get_seed(["torch_seed"], config, int)
        numpy_seed = self._get_seed(["numpy_seed"], config, int)
        random_seed = self._get_seed(["random_seed"], config, int)
        torch.manual_seed(torch_seed)
        torch.cuda.manual_seed_all(torch_seed)
        np.random.seed(numpy_seed)
        random.seed(random_seed)
        self.seeds = {
            "test": test_seed,
            "valid": valid_seed,
            "torch": torch_seed,
            "numpy": numpy_seed,
            "random": random_seed
        }
        self.workers = config["workers"] if "workers" in config and config["workers"] >= 0 else 1
        self.pin_memory = thelper.utils.str2bool(config["pin_memory"]) if "pin_memory" in config else False
        self.drop_last = thelper.utils.str2bool(config["drop_last"]) if "drop_last" in config else False
        self.sampler_type = None
        self.train_sampler, self.valid_sampler, self.test_sampler = None, None, None
        sampler_config = thelper.utils.get_key_def("sampler", config, {})
        if sampler_config:
            sampler_type = thelper.utils.get_key("type", sampler_config)
            self.sampler_type = thelper.utils.import_class(sampler_type)
            self.sampler_params = thelper.utils.get_key_def(["params", "parameters"], sampler_config, {})
            self.sampler_pass_labels = thelper.utils.str2bool(thelper.utils.get_key_def("pass_labels", sampler_config, False))
            self.sampler_pass_labels_param_name = thelper.utils.get_key_def("pass_labels_param_name", sampler_config, "labels")
            self.train_sampler = thelper.utils.str2bool(thelper.utils.get_key_def("apply_train", sampler_config, True))
            self.valid_sampler = thelper.utils.str2bool(thelper.utils.get_key_def("apply_valid", sampler_config, False))
            self.test_sampler = thelper.utils.str2bool(thelper.utils.get_key_def("apply_test", sampler_config, False))
        train_augs_targets = ["augments", "trainvalid_augments", "train_augments"]
        valid_augs_targets = ["augments", "trainvalid_augments", "eval_augments", "validtest_augments", "valid_augments"]
        test_augs_targets = ["augments", "eval_augments", "validtest_augments", "test_augments"]
        self.train_augments, self.train_augments_append = self._get_augments(train_augs_targets, "train", config)
        self.valid_augments, self.valid_augments_append = self._get_augments(valid_augs_targets, "valid", config)
        self.test_augments, self.test_augments_append = self._get_augments(test_augs_targets, "test", config)
        self.base_transforms = None
        if "base_transforms" in config and config["base_transforms"]:
            self.base_transforms = thelper.transforms.load_transforms(config["base_transforms"])
        self.train_split = self._get_ratios_split("train", config)
        self.valid_split = self._get_ratios_split("valid", config)
        self.test_split = self._get_ratios_split("test", config)
        if not self.train_split and not self.valid_split and not self.test_split:
            raise AssertionError("data config must define a split for at least one loader type (train/valid/test)")
        self.total_usage = Counter(self.train_split) + Counter(self.valid_split) + Counter(self.test_split)
        self.skip_split_norm = thelper.utils.str2bool(thelper.utils.get_key_def(
            ["skip_norm", "skip_split_norm"], config, False))
        self.skip_class_balancing = thelper.utils.str2bool(thelper.utils.get_key_def(
            ["skip_balancing", "skip_class_balancing", "skip_rebalancing", "skip_class_rebalancing"], config, False))
        for name, usage in self.total_usage.items():
            if usage != 1:
                normalize_ratios = None
                assert usage >= 0
                if 0 < usage < 1 and not math.isclose(usage, 1) and not self.skip_split_norm:
                    query_msg = f"dataset split for {name} has a ratio sum less than 1; do you want to normalize the split?\n\t("
                    query_msg += f"train={self.train_split[name] if name in self.train_split else 0:.03f}, "
                    query_msg += f"valid={self.valid_split[name] if name in self.valid_split else 0:.03f}, "
                    query_msg += f"test={self.test_split[name] if name in self.test_split else 0:.03f})\n"
                    query_msg += "\t\tto\n\t("
                    query_msg += f"train={self.train_split[name] / usage if name in self.train_split else 0:.03f}, "
                    query_msg += f"valid={self.valid_split[name] / usage if name in self.valid_split else 0:.03f}, "
                    query_msg += f"test={self.test_split[name] / usage if name in self.test_split else 0:.03f})"
                    normalize_ratios = thelper.utils.query_yes_no(query_msg, bypass="n")
                if (normalize_ratios or usage > 1) and usage > 0:
                    if usage > 1:
                        logger.warning("dataset split for '%s' sums to more than 1; will normalize..." % name)
                    if name in self.train_split:
                        self.train_split[name] /= usage
                    if name in self.valid_split:
                        self.valid_split[name] /= usage
                    if name in self.test_split:
                        self.test_split[name] /= usage
        self.skip_verif = thelper.utils.str2bool(config["skip_verif"]) if "skip_verif" in config else True
        logger.debug(f"loaders will use batch sizes:" +
                     (f"\n\ttrain = {self.train_batch_size}" if self.train_split else "") +
                     (f"\n\tvalid = {self.valid_batch_size}" if self.valid_split else "") +
                     (f"\n\ttest = {self.test_batch_size}" if self.test_split else ""))
        logger.debug("samplers will use scaling factors:" +
                     (f"\n\ttrain = {self.train_scale}" if self.train_split else "") +
                     (f"\n\tvalid = {self.valid_scale}" if self.valid_split else "") +
                     (f"\n\ttest = {self.test_scale}" if self.test_split else ""))
        if self.drop_last:
            logger.debug("loaders will drop last batch if sample count not multiple of batch size")
        if sampler_config:
            logger.debug("will use sampler with type '%s' and config : %s" % (str(self.sampler_type), str(self.sampler_params)))
            logger.debug("global sampler will be applied as: %s" % str([self.train_sampler, self.valid_sampler, self.test_sampler]))
        if self.base_transforms:
            logger.debug("base transforms: %s" % str(self.base_transforms))

    @staticmethod
    def _get_collate_fn(val):
        if val is torch.utils.data.dataloader.default_collate or val is default_collate or callable(val):
            return val
        if isinstance(val, dict):
            collate_fn_type = thelper.utils.get_key("type", val)
            collate_fn_params = thelper.utils.get_key_def(["params", "kwargs"], val, None)
            return thelper.utils.import_function(collate_fn_type, collate_fn_params)
        elif isinstance(val, str):
            return thelper.utils.import_function(val)
        else:
            raise AssertionError("unexpected collate val type")

    @staticmethod
    def _get_seed(prefixes, config, stype):
        key = None
        for prefix in prefixes:
            if prefix in config:
                key = prefix
                break
        if key is not None:
            if not isinstance(config[key], stype):
                raise AssertionError("unexpected value type for field '%s'" % key)
            return config[key]
        seed = np.random.randint(2 ** 16)
        logger.info("setting '%s' to %d" % (prefixes[0], seed))
        return seed

    @staticmethod
    def _get_ratios_split(prefix, config):
        key = prefix + "_split"
        if key not in config or not config[key]:
            return {}
        split = config[key]
        assert not any(ratio < 0 for ratio in split.values())
        return split

    @staticmethod
    def _get_augments(targets, name, config):
        logger.debug("loading %s augments..." % name)
        for target in targets:
            if target in config and config[target]:
                augments, augments_append = thelper.transforms.load_augments(config[target])
                if augments:
                    logger.debug("will %s %s augments: %s" % ("append" if augments_append else "prefix", name, str(augments)))
                return augments, augments_append
        return None, False

    def _get_raw_split(self, indices):
        for name in self.total_usage:
            assert name in indices, f"dataset '{name}' does not exist"
        _indices, train_idxs, valid_idxs, test_idxs = {}, {}, {}, {}
        for name, indices in indices.items():
            _indices[name] = copy.deepcopy(indices)
            train_idxs[name] = []
            valid_idxs[name] = []
            test_idxs[name] = []
        indices = _indices
        shuffle = any([self.train_shuffle, self.valid_shuffle, self.test_shuffle])
        if shuffle:
            np.random.seed(self.seeds["test"])  # test idxs will be picked first, then valid+train
            for idxs in indices.values():
                np.random.shuffle(idxs)
        offsets = dict.fromkeys(self.total_usage, 0)
        for loader_idx, (idxs_map, ratio_map) in enumerate(zip([test_idxs, valid_idxs, train_idxs],
                                                               [self.test_split, self.valid_split, self.train_split])):
            for name in self.total_usage.keys():
                if name in ratio_map:
                    count = int(round(ratio_map[name] * len(indices[name])))
                    if count < 0:
                        raise AssertionError("ratios should be non-negative values!")
                    elif count < 1 and len(indices[name]) > 0:
                        logger.warning("split ratio for '%s' too small, sample set will be empty" % name)
                    begidx = offsets[name]
                    endidx = min(begidx + count, len(indices[name]))
                    idxs_map[name] = indices[name][begidx:endidx]
                    offsets[name] = endidx
            if loader_idx == 0 and shuffle:
                np.random.seed(self.seeds["valid"])  # all test idxs are now picked, reshuffle for train/valid
                for name in self.total_usage.keys():
                    trainvalid_idxs = indices[name][offsets[name]:]
                    np.random.shuffle(trainvalid_idxs)
                    indices[name][offsets[name]:] = trainvalid_idxs
        if shuffle:
            np.random.seed(self.seeds["numpy"])  # back to default random state for future use
        return train_idxs, valid_idxs, test_idxs

    def get_split(self, datasets, task):
        r"""Returns the train/valid/test sample indices split for a given dataset (name-parser) map.

        Note that the returned indices are unique, possibly shuffled, and never duplicated between sets.
        If the samples have a class attribute (i.e. the task is related to classification), the split
        will respect the initial distribution and apply the ratios within the classes themselves. For
        example, consider a dataset of three classes (:math:`A`, :math:`B`, and :math:`C`) that contains
        100 samples such as:

        .. math::

            |A| = 50,\;|B| = 30,\;|C| = 20

        If we require a 80%-10%-10% ratio distribution for the training, validation, and test loaders
        respectively, the resulting split will contain the following sample counts:

        .. math::

                \text{training loader} = {40A + 24B + 16C}

        .. math::

                \text{validation loader} = {5A + 3B + 2C}

        .. math::

                \text{test loader} = {5A + 3B + 2C}

        Args:
            datasets: the map of datasets to split, where each has a name (key) and a parser (value).
            task: a task object that should be compatible with all provided datasets (can be ``None``).

        Returns:
            A three-element tuple containing the maps of the training, validation, and test sets
            respectively. These maps associate dataset names to a list of sample indices.
        """
        dataset_sizes = {}
        must_split = {}
        global_size = 0
        for dataset_name, dataset in datasets.items():
            assert isinstance(dataset, thelper.data.Dataset) or isinstance(dataset, thelper.data.ExternalDataset), \
                f"unexpected dataset type for '{dataset_name}'"
            dataset_sizes[dataset_name] = len(dataset)
            global_size += dataset_sizes[dataset_name]
            # if a single dataset is used in more than a single loader, we cannot skip the rebalancing below
            must_split[dataset_name] = sum([dataset_name in split for split in
                                            [self.train_split, self.valid_split, self.test_split]]) > 1
        global_size = sum(len(dataset) for dataset in datasets.values())
        logger.info("splitting datasets with parsed sizes = %s" % str(dataset_sizes))
        must_split = any(must_split.values())
        if task is not None and isinstance(task, thelper.tasks.Classification) and not self.skip_class_balancing and must_split:
            # note: with current impl, all class sets will be shuffled the same way... (shouldnt matter, right?)
            logger.debug("will split evenly over %d classes..." % len(task.class_names))
            unset_class_key = "<unset>"
            global_class_names = task.class_names + [unset_class_key]  # extra name added for unlabeled samples (if needed!)
            sample_maps = {}
            for dataset_name, dataset in datasets.items():
                if isinstance(dataset, thelper.data.ExternalDataset):
                    if hasattr(dataset.samples, "samples") and isinstance(dataset.samples.samples, list):
                        sample_maps[dataset_name] = task.get_class_sample_map(dataset.samples.samples, unset_class_key)
                    else:
                        logger.warning(f"must fully parse the external dataset '{dataset_name}' for balanced intra-class shuffling;" +
                                       " this might take a while!\n\t...consider making a dataset interface that can return labels" +
                                       " only, it would greatly speed up the analysis of class distributions\n\t...you could also" +
                                       " add the 'skip_class_balancing' flag to your data configuration to skip this rebalancing")
                        label_key = task.gt_key
                        # to allow glitch-less tqdm printing after latest logger output
                        sys.stdout.flush()
                        sys.stderr.flush()
                        time.sleep(0.01)
                        samples = []
                        for sample in tqdm.tqdm(dataset):
                            assert label_key in sample, f"could not find label key ('{label_key}') in sample dict"
                            samples.append({label_key: sample[label_key]})
                        sample_maps[dataset_name] = task.get_class_sample_map(samples, unset_class_key)
                elif isinstance(dataset, thelper.data.Dataset):
                    sample_maps[dataset_name] = task.get_class_sample_map(dataset.samples, unset_class_key)
            train_idxs, valid_idxs, test_idxs = {}, {}, {}
            for class_name in global_class_names:
                curr_class_samples, curr_class_size = {}, {}
                for dataset_name in datasets:
                    class_samples = sample_maps[dataset_name][class_name] if class_name in sample_maps[dataset_name] else []
                    samples_pairs = list(zip(class_samples, [class_name] * len(class_samples)))
                    curr_class_samples[dataset_name] = samples_pairs
                    curr_class_size[dataset_name] = len(curr_class_samples[dataset_name])
                    logger.debug("dataset '{}' class '{}' sample count: {} ({}% of local, {}% of total)".format(
                        dataset_name,
                        class_name,
                        curr_class_size[dataset_name],
                        int(100 * curr_class_size[dataset_name] / dataset_sizes[dataset_name]),
                        int(100 * curr_class_size[dataset_name] / global_size)))
                class_train_idxs, class_valid_idxs, class_test_idxs = self._get_raw_split(curr_class_samples)
                for idxs_dict_list, class_idxs_dict_list in zip([train_idxs, valid_idxs, test_idxs],
                                                                [class_train_idxs, class_valid_idxs, class_test_idxs]):
                    for dataset_name in datasets:
                        if dataset_name in idxs_dict_list:
                            idxs_dict_list[dataset_name] += class_idxs_dict_list[dataset_name]
                        else:
                            idxs_dict_list[dataset_name] = class_idxs_dict_list[dataset_name]
        else:  # no balancing to be done
            dataset_indices = {}
            for dataset_name in datasets:
                # note: all indices paired with 'None' below as class is ignored; used for compatibility with code above
                dataset_indices[dataset_name] = list(
                    zip(list(range(dataset_sizes[dataset_name])), [None] * dataset_sizes[dataset_name]))
            train_idxs, valid_idxs, test_idxs = self._get_raw_split(dataset_indices)
        return train_idxs, valid_idxs, test_idxs

    def create_loaders(self, datasets, train_idxs, valid_idxs, test_idxs):
        """Returns the data loaders for the train/valid/test sets based on a prior split.

        This function essentially takes the dataset parser interfaces and indices maps, and instantiates
        data loaders that are ready to produce samples for training or evaluation. Note that the dataset
        parsers will be deep-copied in each data loader, meaning that they should ideally not contain a
        persistent loading state or a large buffer.

        Args:
            datasets: the map of dataset parsers, where each has a name (key) and a parser (value).
            train_idxs: training data samples indices map.
            valid_idxs: validation data samples indices map.
            test_idxs: test data samples indices map.

        Returns:
            A three-element tuple containing the training, validation, and test data loaders, respectively.
        """
        loaders = []
        for idxs_map, (augs, augs_append), shuffle, scale, sampler_apply, batch_size, collate_fn \
                in zip([train_idxs, valid_idxs, test_idxs],
                       [(self.train_augments, self.train_augments_append),
                        (self.valid_augments, self.valid_augments_append),
                        (self.test_augments, self.test_augments_append)],
                       [self.train_shuffle, self.valid_shuffle, self.test_shuffle],
                       [self.train_scale, self.valid_scale, self.test_scale],
                       [self.train_sampler, self.valid_sampler, self.test_sampler],
                       [self.train_batch_size, self.valid_batch_size, self.test_batch_size],
                       [self.train_collate_fn, self.valid_collate_fn, self.test_collate_fn]):
            loader_sample_idx_offset = 0
            loader_sample_classes = []
            loader_sample_idxs = []
            loader_datasets = []
            for dataset_name, sample_idxs in idxs_map.items():
                if not sample_idxs:
                    continue
                # todo: investigate need to copy at all?
                if not datasets[dataset_name].deepcopy:
                    dataset = copy.copy(datasets[dataset_name])
                else:
                    dataset = copy.deepcopy(datasets[dataset_name])
                if augs:
                    augs_copy = copy.deepcopy(augs)
                    if dataset.transforms is not None:
                        if augs_append:
                            dataset.transforms = thelper.transforms.Compose([dataset.transforms, augs_copy])
                        else:
                            dataset.transforms = thelper.transforms.Compose([augs_copy, dataset.transforms])
                    else:
                        dataset.transforms = augs_copy
                for sample_idx_idx in range(len(sample_idxs)):
                    # values were paired in tuples earlier, 0=idx, 1=label
                    loader_sample_idxs.append(sample_idxs[sample_idx_idx][0] + loader_sample_idx_offset)
                    loader_sample_classes.append(sample_idxs[sample_idx_idx][1])
                loader_sample_idx_offset += len(dataset)
                loader_datasets.append(dataset)
            if len(loader_datasets) > 0:
                dataset = torch.utils.data.ConcatDataset(loader_datasets) if len(loader_datasets) > 1 else loader_datasets[0]
                if self.sampler_type is not None and sampler_apply:
                    sampler_params = {**self.sampler_params}
                    if self.sampler_pass_labels:
                        sampler_params[self.sampler_pass_labels_param_name] = loader_sample_classes
                    sampler_sig = inspect.signature(self.sampler_type)
                    if "seeds" in sampler_sig.parameters:
                        sampler_params["seeds"] = self.seeds
                    if "scale" in sampler_sig.parameters:
                        sampler_params["scale"] = scale
                    else:
                        assert scale == 1.0, f"could not apply scale factor to sample with type '{str(self.sampler_type)}'"
                    sampler = self.sampler_type(loader_sample_idxs, **sampler_params)
                else:
                    if shuffle:
                        sampler = thelper.data.SubsetRandomSampler(loader_sample_idxs, seeds=self.seeds, scale=scale)
                    else:
                        assert scale == 1.0, "sequential sampler currently does not handle scale changes (turn on shuffling)"
                        sampler = thelper.data.SubsetSequentialSampler(loader_sample_idxs)
                assert hasattr(sampler, "__len__")
                assert batch_size > 0
                loaders.append(DataLoader(dataset=dataset, batch_size=batch_size, sampler=sampler,
                                          num_workers=self.workers, collate_fn=collate_fn,
                                          pin_memory=self.pin_memory, drop_last=self.drop_last,
                                          seeds=self.seeds))
            else:
                loaders.append(None)
        train_loader, valid_loader, test_loader = loaders
        logger.info("initialized loaders with batch counts:" +
                    (f"\n\ttrain = {len(train_loader)}" if train_loader else "") +
                    (f"\n\tvalid = {len(valid_loader)}" if valid_loader else "") +
                    (f"\n\ttest = {len(test_loader)}" if test_loader else ""))
        return train_loader, valid_loader, test_loader

    def get_base_transforms(self):
        """Returns the (global) sample transformation operations parsed in the data configuration."""
        return self.base_transforms



import torch


class DenseBlock(torch.nn.Module):
    def __init__(self, input_size, output_size, bias=True, activation='relu', norm='batch'):
        super(DenseBlock, self).__init__()
        self.fc = torch.nn.Linear(input_size, output_size, bias=bias)
        self.norm = norm
        if self.norm == 'batch':
            self.bn = torch.nn.BatchNorm1d(output_size)
        elif self.norm == 'instance':
            self.bn = torch.nn.InstanceNorm1d(output_size)
        self.activation = activation
        if self.activation == 'relu':
            self.act = torch.nn.ReLU(True)
        elif self.activation == 'prelu':
            self.act = torch.nn.PReLU()
        elif self.activation == 'lrelu':
            self.act = torch.nn.LeakyReLU(0.2, True)
        elif self.activation == 'tanh':
            self.act = torch.nn.Tanh()
        elif self.activation == 'sigmoid':
            self.act = torch.nn.Sigmoid()

    def forward(self, x):
        if self.norm is not None:
            out = self.bn(self.fc(x))
        else:
            out = self.fc(x)
        if self.activation is not None:
            return self.act(out)
        else:
            return out


class ConvBlock(torch.nn.Module):
    def __init__(self, input_size, output_size, kernel_size=4, stride=2, padding=1,
                 bias=True, activation='relu', norm='batch', groups=1, prelu_params=1):
        super(ConvBlock, self).__init__()
        self.conv = torch.nn.Conv2d(input_size, output_size, kernel_size,
                                    stride, padding, bias=bias, groups=groups)
        self.norm = norm
        if self.norm == 'batch':
            self.bn = torch.nn.BatchNorm2d(output_size)
        elif self.norm == 'instance':
            self.bn = torch.nn.InstanceNorm2d(output_size)
        elif self.norm is None:
            self.bn = None
        else:
            raise(Exception('Bad normalization selection'))
        self.activation = activation
        if self.activation == 'relu':
            self.act = torch.nn.ReLU(True)
        elif self.activation == 'prelu':
            if prelu_params != 1:
                prelu_params = input_size
            self.act = torch.nn.PReLU(num_parameters=prelu_params)
        elif self.activation == 'lrelu':
            self.act = torch.nn.LeakyReLU(0.2, True)
        elif self.activation == 'tanh':
            self.act = torch.nn.Tanh()
        elif self.activation == 'sigmoid':
            self.act = torch.nn.Sigmoid()
        elif self.activation is None:
            self.act = None
        else:
            raise(Exception('Bad activation selection'))
        self.forward = self.forward_bn_act

    def forward_bn_act(self, x):
        x = self.conv(x)
        if self.bn is not None:
            x = self.bn(x)
        if self.act is not None:
            x = self.act(x)
        return x

    def forward_act_bn(self, x):
        x = self.conv(x)
        if self.act is not None:
            x = self.act(x)
        if self.bn is not None:
            x = self.bn(x)
        return x


class DeconvBlock(torch.nn.Module):
    def __init__(self, input_size, output_size, kernel_size=4, stride=2, padding=1,
                 bias=True, activation='relu', norm='batch'):
        super(DeconvBlock, self).__init__()
        self.deconv = torch.nn.ConvTranspose2d(input_size, output_size, kernel_size,
                                               stride, padding, bias=bias)
        self.norm = norm
        if self.norm == 'batch':
            self.bn = torch.nn.BatchNorm2d(output_size)
        elif self.norm == 'instance':
            self.bn = torch.nn.InstanceNorm2d(output_size)
        self.activation = activation
        if self.activation == 'relu':
            self.act = torch.nn.ReLU(True)
        elif self.activation == 'prelu':
            self.act = torch.nn.PReLU()
        elif self.activation == 'lrelu':
            self.act = torch.nn.LeakyReLU(0.2, True)
        elif self.activation == 'tanh':
            self.act = torch.nn.Tanh()
        elif self.activation == 'sigmoid':
            self.act = torch.nn.Sigmoid()

    def forward(self, x):
        if self.norm is not None:
            out = self.bn(self.deconv(x))
        else:
            out = self.deconv(x)
        if self.activation is not None:
            return self.act(out)
        else:
            return out


class ResNetBlock(torch.nn.Module):
    def __init__(self, num_filter, kernel_size=3, stride=1, padding=1,
                 bias=True, activation='relu', norm='batch'):
        super(ResNetBlock, self).__init__()
        self.conv1 = torch.nn.Conv2d(num_filter, num_filter, kernel_size,
                                     stride, padding, bias=bias)
        self.conv2 = torch.nn.Conv2d(num_filter, num_filter, kernel_size,
                                     stride, padding, bias=bias)
        self.norm = norm
        if self.norm == 'batch':
            self.bn = torch.nn.BatchNorm2d(num_filter)
        elif norm == 'instance':
            self.bn = torch.nn.InstanceNorm2d(num_filter)
        self.activation = activation
        if self.activation == 'relu':
            self.act = torch.nn.ReLU(True)
        elif self.activation == 'prelu':
            self.act = torch.nn.PReLU()
        elif self.activation == 'lrelu':
            self.act = torch.nn.LeakyReLU(0.2, True)
        elif self.activation == 'tanh':
            self.act = torch.nn.Tanh()
        elif self.activation == 'sigmoid':
            self.act = torch.nn.Sigmoid()

    def forward(self, x):
        residual = x
        if self.norm is not None:
            out = self.bn(self.conv1(x))
        else:
            out = self.conv1(x)
        if self.activation is not None:
            out = self.act(out)
        if self.norm is not None:
            out = self.bn(self.conv2(out))
        else:
            out = self.conv2(out)
        out = torch.add(out, residual)
        return out


class PSBlock(torch.nn.Module):
    def __init__(self, input_size, output_size, scale_factor, kernel_size=3,
                 stride=1, padding=1, bias=True, activation='relu', norm='batch'):
        super(PSBlock, self).__init__()
        self.conv = torch.nn.Conv2d(input_size, output_size * scale_factor**2, kernel_size, stride, padding, bias=bias)
        self.ps = torch.nn.PixelShuffle(scale_factor)
        self.norm = norm
        if self.norm == 'batch':
            self.bn = torch.nn.BatchNorm2d(output_size)
        elif norm == 'instance':
            self.bn = torch.nn.InstanceNorm2d(output_size)
        self.activation = activation
        if self.activation == 'relu':
            self.act = torch.nn.ReLU(True)
        elif self.activation == 'prelu':
            self.act = torch.nn.PReLU()
        elif self.activation == 'lrelu':
            self.act = torch.nn.LeakyReLU(0.2, True)
        elif self.activation == 'tanh':
            self.act = torch.nn.Tanh()
        elif self.activation == 'sigmoid':
            self.act = torch.nn.Sigmoid()

    def forward(self, x):
        if self.norm is not None:
            out = self.bn(self.ps(self.conv(x)))
        else:
            out = self.ps(self.conv(x))
        if self.activation is not None:
            out = self.act(out)
        return out


class Upsample2xBlock(torch.nn.Module):
    def __init__(self, input_size, output_size, bias=True, upsample='deconv', activation='relu', norm='batch'):
        super(Upsample2xBlock, self).__init__()
        scale_factor = 2
        # 1. Deconvolution (Transposed convolution)
        if upsample == 'deconv':
            self.upsample = DeconvBlock(input_size, output_size,
                                        kernel_size=4, stride=2, padding=1,
                                        bias=bias, activation=activation, norm=norm)
        # 2. Sub-pixel convolution (Pixel shuffler)
        elif upsample == 'ps':
            self.upsample = PSBlock(input_size, output_size, scale_factor=scale_factor,
                                    bias=bias, activation=activation, norm=norm)
        # 3. Resize and Convolution
        elif upsample == 'rnc':
            self.upsample = torch.nn.Sequential(
                torch.nn.Upsample(scale_factor=scale_factor, mode='nearest'),
                ConvBlock(input_size, output_size,
                          kernel_size=3, stride=1, padding=1,
                          bias=bias, activation=activation, norm=norm)
            )

    def forward(self, x):
        out = self.upsample(x)
        return out


def weights_init_kaiming(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
        torch.nn.init.kaiming_normal(m.weight)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('Conv2d') != -1:
        torch.nn.init.kaiming_normal(m.weight)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('ConvTranspose2d') != -1:
        torch.nn.init.kaiming_normal(m.weight)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('Norm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        if m.bias is not None:
            m.bias.data.zero_()


def weights_init_xavier(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
        torch.nn.init.xavier_uniform(m.weight)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('Conv2d') != -1:
        torch.nn.init.xavier_uniform(m.weight)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('ConvTranspose2d') != -1:
        torch.nn.init.xavier_uniform(m.weight)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('Norm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        if m.bias is not None:
            m.bias.data.zero_()


def shave(imgs, border_size=0):
    size = list(imgs.shape)
    if len(size) == 4:
        shave_imgs = torch.FloatTensor(size[0], size[1], size[2] - border_size * 2, size[3] - border_size * 2)
        for i, img in enumerate(imgs):
            shave_imgs[i, :, :, :] = img[:, border_size:-border_size, border_size:-border_size]
        return shave_imgs
    else:
        return imgs[:, border_size:-border_size, border_size:-border_size]
